#include <OneWire.h>
#include <DallasTemperature.h>
#include <SoftwareSerial.h>
 
#define DEBUG true

#define SOIL_MOISTURE_SENSOR A1
#define WATER_LEVEL_SENSOR A2
#define MOTOR 11
#define ONE_WIRE_BUS A0

OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);

const int soilMoisturePower = 13;
const int waterLevelPower = 12;

const int upLevelMoisture = 950;

const int downLevelWater = 670;

boolean isAutomatic = true;
boolean isCelsius = true;
int wateringIntensity = 2000;

char  strArray[3];
 
SoftwareSerial esp8266(2,3); // Make RX Arduino line on pin 2, make TX Arduino line on pin 3.
                             // This means that you need to connect the TX line from the ESP to the Arduino's pin 2
                             // and the RX line from the ESP to the Arduino's pin 3

void setup() {
  pinMode(SOIL_MOISTURE_SENSOR, INPUT);
  pinMode(WATER_LEVEL_SENSOR,INPUT);
  pinMode(MOTOR, OUTPUT);
  digitalWrite(MOTOR, HIGH);
  pinMode(soilMoisturePower, OUTPUT);
  digitalWrite(soilMoisturePower, LOW);
  pinMode(waterLevelPower, OUTPUT);
  digitalWrite(waterLevelPower, LOW);
  
  sensors.begin();
  Serial.begin(9600);
  esp8266.begin(115200); // Your esp's baud rate might be different
   
  sendCommand("AT+RST\r\n",2000,DEBUG); // Reset module
  sendCommand("AT+CWMODE=2\r\n",2000,DEBUG); // Configure as access point
  sendCommand("AT+CIFSR\r\n",2000,DEBUG); // Get ip address
  sendCommand("AT+CIPMUX=1\r\n",2000,DEBUG); // Configure for multiple connections
  sendCommand("AT+CIPSERVER=1,80\r\n",2000,DEBUG); // Turn on server on port 80
  
  Serial.println("Server Ready");
}

void loop() {
  
  sensors.requestTemperatures();
  float temperature = sensors.getTempCByIndex(0);
  delay(50);
  float soilMoistureValue = readMoistureSensor();
  delay(50);
  float waterLevelValue = readWaterSensor();

  // Check if the ESP is sending a message 
  if(esp8266.available()) 
  {   
    if(esp8266.find("+IPD,"))
    {
     // Wait for the serial buffer to fill up (read all the serial data)
     delay(1000); 

     // Get the connection id so that we can then disconnect.
     // Subtract 48 because the read() function returns 
     // the ASCII decimal value and 0 (the first decimal number) starts at 48
     int connectionId = esp8266.read()-48;
     
     // Build the string that is sent back to the device that is sending the request
     String content = "";

     // Build the string, that stores a message sent from the device to the ESP
     String request = "";

     // advance cursor to "/?"
     esp8266.find("/?");

     // Read enough characters to encompass the request sent from the device
     for (int i = 0; i < 30; i++) {
        char c = esp8266.read();
        request += String(c);
     }

     // Apply to the Arduino board new settings values, if parameter's key equals "settings"
     if (request.startsWith("settings")){
        
        split(request);
        
        isAutomatic = (strArray[0] == 't') ? true : false;
        wateringIntensity = (strArray[1] == 'l') ? 1000 : ((strArray[1] == 'm') ? 2000 : 3000);
        isCelsius = (strArray[2] == 'c') ? true : false;

        content = "OK";

     // Return a particular sensor value in percentage as the response, if parameter's key equals "sensor"
     } else if (request.startsWith("sensor")) {
      
        int index = request.indexOf("=");
        
        switch(request[index + 1]) {
          case 'w':
            content = String(min(mapFloat(waterLevelValue, 650, 690, 0, 100), 100));
            break;
          case 't':
            if (!isCelsius) {
              temperature = temperature * 9 / 5 + 32; // Change the temperature value in celsius to fahrenheit
            }
            content = String(temperature);
            break;
          case 'm':
            content = String(min(mapFloat(soilMoistureValue, 1023, 890, 0, 100), 100));
            break;
        }
     // Water the plant, if parameter's key equals "pourOn"
     } else if (request.startsWith("pourOn")) {
        
        content = "OK";
      
        digitalWrite(MOTOR, LOW);
        delay(wateringIntensity);
        digitalWrite(MOTOR, HIGH);
      
     }

     sendHTTPResponse(connectionId,content);
     
     String closeCommand = "AT+CIPCLOSE="; 
     closeCommand += connectionId; // Append connection id
     closeCommand += "\r\n";
     
     sendCommand(closeCommand,1000,DEBUG);
    }
  }

  if (waterLevelValue > downLevelWater) {
     if (isAutomatic) {
        if (soilMoistureValue > upLevelMoisture) {
            digitalWrite(MOTOR, LOW);
            delay(wateringIntensity);
            digitalWrite(MOTOR, HIGH);
        }
     } else digitalWrite(MOTOR, HIGH);
  } else digitalWrite(MOTOR, HIGH);
   
  delay(1000);
}


/*
* Name: mapFloat
* Description: Re-maps a number from one range to another
* Params: num - the number to map; fromMin - the lower bound of the value’s current range; 
  fromMax - the upper bound of the value’s current range; toMin - the lower bound of the value’s target range; 
  toMax - the upper bound of the value’s target range
* Returns: The mapped value
*/
float mapFloat(float num, float fromMin, float fromMax, float toMin, float toMax) {
  float fromDiff = fromMax - fromMin;
  float toDiff = toMax - toMin;
  float result = (num - fromMin) * toDiff / fromDiff + toMin;
  return result;
}


/*
* Name: split
* Description: Splits the request string by the "=" delimiter, 
  storing the first characters of the parameters' values 
  (this function is used ONLY for "settings" parameter key)
* Params: The request string sent by the device
* Returns: Nothing, stores characters in the global strArray
*/
void split(String str) {
  
  int index = str.indexOf("=") + 1;
  
  for (int i = 0; i < 3; i++) {
    strArray[i] = str.charAt(index);
    index = str.indexOf("&", index) + 1;
  }
  
}


/*
* Name: readMoistureSensor
* Description: Reads the value from the moisture sensor.
  This function is used to extend sensor life longevity by applying 5V voltage 
  only when it is required to read the sensor value
* Returns: The moisture sensor value
*/
int readMoistureSensor() {
  digitalWrite(soilMoisturePower, HIGH);
  delay(10);
  int val = analogRead(SOIL_MOISTURE_SENSOR);
  digitalWrite(soilMoisturePower, LOW);
  return val;
}


/*
* Name: readWaterSensor
* Description: Reads the value from the water level sensor. 
  This function is used to extend sensor life longevity by applying 5V voltage 
  only when it is required to read the sensor value
* Returns: The water level sensor value
*/
int readWaterSensor() {
  digitalWrite(waterLevelPower, HIGH);
  delay(10);
  int val = analogRead(WATER_LEVEL_SENSOR);
  digitalWrite(waterLevelPower, LOW);
  return val;
}


/*
* Name: sendData
* Description: Function used to send data to ESP8266
* Params: command - the data/command to send; timeout - the time to wait for a response; 
  debug - print to Serial window?(true = yes, false = no)
* Returns: The response from the ESP8266 (if there is a reponse)
*/
String sendData(String command, const int timeout, boolean debug)
{
    String response = "";
    
    int dataSize = command.length();
    char data[dataSize];
    command.toCharArray(data,dataSize);
           
    esp8266.write(data,dataSize); // Send the characters read to the ESP8266
    if(debug)
    {
      Serial.println("\r\n====== HTTP Response From Arduino ======");
      Serial.write(data,dataSize);
      Serial.println("\r\n========================================");
    }
    
    long int time = millis();
    
    while( (time+timeout) > millis())
    {
      while(esp8266.available())
      {

        // The ESP has data so display its output to the serial window 
        char c = esp8266.read(); // Read the next character.
        response += c;
      }  
    }
    
    if(debug)
    {
      Serial.print(response);
    }
    
    return response;
}


/*
* Name: sendHTTPResponse
* Description: Function that sends HTTP 200, HTML UTF-8 response
*/
void sendHTTPResponse(int connectionId, String content)
{
     // Build HTTP response
     String httpResponse;
     String httpHeader;

     // HTTP Header
     httpHeader = "HTTP/1.1 200 OK\r\nContent-Type: text/html; charset=UTF-8\r\n"; 
     httpHeader += "Content-Length: ";
     httpHeader += content.length();
     httpHeader += "\r\n";
     httpHeader += "Connection: close\r\n\r\n";
     // There is a bug in this code: the last character of "content" is not sent.
     httpResponse = httpHeader + content + " "; // I cheated by adding this extra space
     sendCIPData(connectionId,httpResponse);
}


/*
* Name: sendCIPDATA
* Description: sends a CIPSEND=<connectionId>,<data> command
*/
void sendCIPData(int connectionId, String data)
{
   String cipSend = "AT+CIPSEND=";
   cipSend += connectionId;
   cipSend += ",";
   cipSend += data.length();
   cipSend += "\r\n";
   sendCommand(cipSend,1000,DEBUG);
   sendData(data,1000,DEBUG);
}

/*
* Name: sendCommand
* Description: Function used to send data to ESP8266.
* Params: command - the data/command to send; timeout - the time to wait for a response; 
  debug - print to Serial window?(true = yes, false = no)
* Returns: The response from the ESP8266 (if there is a reponse)
*/
String sendCommand(String command, const int timeout, boolean debug)
{
    String response = "";
           
    esp8266.print(command); // send the characters read to the ESP8266
    long int time = millis();
    
    while( (time+timeout) > millis())
    {
      while(esp8266.available())
      {

        // The ESP has data so display its output to the serial window 
        char c = esp8266.read(); // Read the next character
        response += c;
      }  
    }
    
    if(debug)
    {
      Serial.print(response);
    }
    
    return response;
}
